import calendar
import concurrent.futures
import logging
import pathlib
import re
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Iterator, Literal, Sequence, TypeVar

import numpy
import xarray
from dateutil import parser
from dateutil.parser._parser import ParserError
from requests import PreparedRequest
from tqdm import tqdm

from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_USE_THREADS,
)
from copernicusmarine.core_functions.exceptions import WrongDatetimeFormat
from copernicusmarine.versioner import __version__ as copernicusmarine_version

logger = logging.getLogger("copernicusmarine")


def get_unique_filepath(
    filepath: pathlib.Path,
) -> pathlib.Path:
    parent = filepath.parent
    filename = filepath.stem
    extension = filepath.suffix
    counter = 1

    while filepath.exists():
        filepath = parent / (filename + "_(" + str(counter) + ")" + extension)
        counter += 1
    return filepath


def get_unique_directorypath(
    dirpath: pathlib.Path,
) -> pathlib.Path:
    parent = dirpath.parent
    dirname = dirpath.name
    counter = 1

    while dirpath.exists():
        dirpath = parent / (dirname + "_(" + str(counter) + ")")
        counter += 1
    return dirpath


_T = TypeVar("_T")


def next_or_raise_exception(
    iterator: Iterator[_T], exception_to_raise: Exception
) -> _T:
    try:
        return next(iterator)
    except Exception as exception:
        raise exception_to_raise from exception


def construct_url_with_query_params(
    url: str, query_params: dict[str, str]
) -> str | None:
    req = PreparedRequest()
    req.prepare_url(url, query_params)
    return req.url


def construct_query_params_for_marine_data_store_monitoring(
    username: str | None = None,
) -> dict[str, str]:
    query_params = {
        "x-cop-client": "copernicus-marine-toolbox",
        "x-cop-client-version": copernicusmarine_version,
    }
    if username:
        query_params["x-cop-user"] = username
    return query_params


def datetime_parser(date: str | numpy.datetime64) -> datetime:
    if date == "now":
        return datetime.now(tz=timezone.utc)
    try:
        if isinstance(date, numpy.datetime64):
            date = str(date)
        parsed_datetime = parser.parse(
            date, default=datetime(1978, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
        )
        return parsed_datetime
    except ParserError:
        pass
    raise WrongDatetimeFormat(date)


def timestamp_parser(
    timestamp: int | float, unit: Literal["s", "ms"] = "ms"
) -> datetime:
    """
    Convert a timestamp in milliseconds to a datetime object.
    The unit can be changed to seconds by passing "s".
    """
    delta = (
        timedelta(seconds=timestamp)
        if unit == "s"
        else timedelta(milliseconds=timestamp)
    )
    return datetime(1970, 1, 1, tzinfo=timezone.utc) + delta


def datetime_to_timestamp(
    date: datetime, unit: Literal["s", "ms"] = "ms"
) -> int | float:
    """
    Should be Windows compatible for datetime before 1970
    Returns a timestamp corresponding to the seconds or
    milliseconds since the epoch (January 1, 1970, 00:00:00 UTC).
    """
    return calendar.timegm(date.timetuple()) * (1000 if unit == "ms" else 1)


def timestamp_or_datestring_to_datetime(
    date: str | int | float | numpy.datetime64,
) -> datetime:
    if isinstance(date, int) or isinstance(date, float):
        return timestamp_parser(date)
    else:
        return datetime_parser(date)


def datetime_to_isoformat(
    date: datetime,
) -> str:
    """
    Convert a datetime object to ISO 8601 format.
    For consistency we want to return a timezone-aware datetime.

    Example: 2023-11-25T00:00:00+00:00
    or 2023-11-25T00:00:00Z (we prefer the latter)
    """
    return date.isoformat().replace("+00:00", "Z")


def add_copernicusmarine_version_in_dataset_attributes(
    dataset: xarray.Dataset,
) -> xarray.Dataset:
    dataset.attrs["copernicusmarine_version"] = copernicusmarine_version
    return dataset


# From: https://stackoverflow.com/a/46144596/20983727
def run_concurrently(
    func: Callable[..., _T],
    function_arguments: Sequence[tuple[Any, ...]],
    max_concurrent_requests: int,
    tdqm_bar_configuration: dict = {},
) -> list[_T]:
    out = []
    with tqdm(
        total=len(function_arguments),
        **tdqm_bar_configuration,
    ) as pbar:
        if max_concurrent_requests <= 0 or not COPERNICUSMARINE_USE_THREADS:
            for function_argument in function_arguments:
                out.append(func(*function_argument))
                pbar.update(1)
        else:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=max_concurrent_requests
            ) as executor:
                future_to_url = (
                    executor.submit(func, *function_argument)
                    for function_argument in function_arguments
                )
                for future in concurrent.futures.as_completed(future_to_url):
                    data = future.result()
                    out.append(data)
                    pbar.update(1)
    return out


def run_multiprocessors(
    func: Callable[..., _T],
    function_arguments: Sequence[tuple[Any, ...]],
    max_concurrent_requests: int,
    tdqm_bar_configuration: dict = {},
) -> list[_T]:
    out = []
    with tqdm(
        total=len(function_arguments),
        **tdqm_bar_configuration,
    ) as pbar:
        with concurrent.futures.ProcessPoolExecutor(
            max_workers=max_concurrent_requests,
        ) as executor:
            try:
                future_to_url = (
                    executor.submit(func, *function_argument)
                    for function_argument in function_arguments
                )
                for future in concurrent.futures.as_completed(future_to_url):
                    data = future.result()
                    out.append(data)
                    pbar.update(1)
                return out
            except KeyboardInterrupt:
                logger.warning(
                    "Terminating all parallel processes, this could take a while..."
                )
                executor.shutdown(wait=False, cancel_futures=True)
                raise


# Example data_path
# https://s3.waw3-1.cloudferro.com/mdl-native-01/native/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m_202105
# https://s3.region.cloudferro.com/bucket/arco/product/dataset/geoChunked.zarr
# https://s3.region.cloudferro.com:443/bucket/arco/product/dataset/geoChunked.zarr
def parse_access_dataset_url(
    data_path: str, only_dataset_root_path: bool = False
) -> tuple[str, str, str]:

    match = re.search(
        r"^(http|https):\/\/([\w\-\.]+)(:[\d]+)?(\/.*)", data_path
    )
    if match:
        endpoint_url = match.group(1) + "://" + match.group(2)
        full_path = match.group(4)
        segments = full_path.split("/")
        bucket = segments[1]
        path = (
            "/".join(segments[2:])
            if not only_dataset_root_path
            else "/".join(segments[2:5]) + "/"
        )
        return endpoint_url, bucket, path
    else:
        raise ValueError(f"Invalid data path: {data_path}")


def create_custom_query_function(username: str | None) -> Callable:
    def _add_custom_query_param(params, context, **kwargs):
        """
        Add custom query params for MDS's Monitoring
        """
        params["url"] = construct_url_with_query_params(
            params["url"],
            construct_query_params_for_marine_data_store_monitoring(username),
        )

    return _add_custom_query_param


def human_readable_size(size_mb: numpy.float64 | float) -> str:
    units = ["B", "KB", "MB", "GB", "TB", "PB"]
    size_bytes = size_mb * 1024 * 1024
    index = 0

    while size_bytes >= 1024 and index < len(units) - 1:
        size_bytes /= 1024.0
        index += 1

    return f"{size_bytes:.2f} {units[index]}"



import pathlib
from datetime import datetime

import pandas as pd

from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
    deprecated_python_option,
)
from copernicusmarine.core_functions.exceptions import (
    MutuallyExclusiveArguments,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_COORDINATES_SELECTION_METHOD,
    DEFAULT_VERTICAL_AXIS,
    CoordinatesSelectionMethod,
    FileFormat,
    ResponseSubset,
    VerticalAxis,
)
from copernicusmarine.core_functions.request_structure import (
    create_subset_request,
)
from copernicusmarine.core_functions.subset import subset_function
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def subset(
    dataset_id: str | None = None,
    dataset_version: str | None = None,
    dataset_part: str | None = None,
    username: str | None = None,
    password: str | None = None,
    variables: list[str] | None = None,
    minimum_longitude: float | None = None,
    maximum_longitude: float | None = None,
    minimum_latitude: float | None = None,
    maximum_latitude: float | None = None,
    minimum_depth: float | None = None,
    maximum_depth: float | None = None,
    vertical_axis: VerticalAxis = DEFAULT_VERTICAL_AXIS,  # noqa
    start_datetime: datetime | pd.Timestamp | str | None = None,
    end_datetime: datetime | pd.Timestamp | str | None = None,
    minimum_x: float | None = None,
    maximum_x: float | None = None,
    minimum_y: float | None = None,
    maximum_y: float | None = None,
    coordinates_selection_method: CoordinatesSelectionMethod = (
        DEFAULT_COORDINATES_SELECTION_METHOD
    ),
    output_filename: str | None = None,
    file_format: FileFormat | None = None,
    service: str | None = None,
    request_file: pathlib.Path | str | None = None,
    output_directory: pathlib.Path | str | None = None,
    credentials_file: pathlib.Path | str | None = None,
    motu_api_request: str | None = None,
    overwrite: bool = False,
    skip_existing: bool = False,
    dry_run: bool = False,
    disable_progress_bar: bool = False,
    staging: bool = False,
    netcdf_compression_level: int = 0,
    netcdf3_compatible: bool = False,
    chunk_size_limit: int = -1,
    raise_if_updating: bool = False,
    platform_ids: list[str] | None = None,
) -> ResponseSubset:
    """
    Extract a subset of data from a specified dataset using given parameters.

    The datasetID is required and can be found via the ``describe`` command.

    Parameters
    ----------
    dataset_id : str, optional
        The datasetID, required either as an argument or in the request_file option.
    dataset_version : str, optional
        Force the selection of a specific dataset version.
    dataset_part : str, optional
        Force the selection of a specific dataset part.
    username : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_USERNAME, then search for a credentials file, else ask for user input. See also :func:`~copernicusmarine.login`
    password : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_PASSWORD, then search for a credentials file, else ask for user input. See also :func:`~copernicusmarine.login`
    variables : list[str], optional
        List of variable names to extract.
    minimum_longitude : float, optional
        Minimum longitude for the subset. The value will be transposed to the interval [-180; 360[.
    maximum_longitude : float, optional
        Maximum longitude for the subset. The value will be transposed to the interval [-180; 360[.
    minimum_latitude : float, optional
        Minimum latitude for the subset. Requires a float from -90 degrees to +90.
    maximum_latitude : float, optional
        Maximum latitude for the subset. Requires a float from -90 degrees to +90.
    minimum_x : float, optional
        Minimum x-axis value for the subset. The units are considered in length (m, 100km...).
    maximum_x : float, optional
        Maximum x-axis value for the subset. The units are considered in length (m, 100km...).
    minimum_y : float, optional
        Minimum y-axis value for the subset. The units are considered in length (m, 100km...).
    maximum_y : float, optional
        Maximum y-axis value for the subset. The units are considered in length (m, 100km...).
    minimum_depth : float, optional
        Minimum depth for the subset.
    maximum_depth : float, optional
        Maximum depth for the subset.
    vertical_axis : str, optional
        Consolidate the vertical dimension (the z-axis) as requested: depth with descending positive values, elevation with ascending positive values. Default is depth.
    start_datetime : datetime | str, optional
        The start datetime of the temporal subset. Supports common format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html).
    end_datetime : datetime | str, optional
        The end datetime of the temporal subset. Supports common format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html).
    coordinates_selection_method : str, optional
        If ``inside``, the selection retrieved will be inside the requested range. If ``strict-inside``, the selection retrieved will be inside the requested range, and an error will be raised if the values don't exist. If ``nearest``, the extremes closest to the requested values will be returned. If ``outside``, the extremes will be taken to contain all the requested interval. The methods ``inside``, ``nearest`` and ``outside`` will display a warning if the request is out of bounds.
    output_directory : pathlib.Path | str, optional
        The destination folder for the downloaded files. Default is the current directory.
    credentials_file : pathlib.Path | str, optional
        Path to a credentials file if not in its default directory (``$HOME/.copernicusmarine``). Accepts .copernicusmarine-credentials / .netrc or _netrc / motuclient-python.ini files.
    output_filename : str, optional
        Save the downloaded data with the given file name (under the output directory). Extension is optional and will be added if not set. Extension takes priority over the file format option if both are set.
    file_format : str, optional
        Format of the downloaded dataset. If not set or set to ``None``, defaults to NetCDF '.nc' for gridded datasets and to CSV '.csv' for sparse datasets. Output filename extension takes priority over this option if both are set.
    overwrite : bool, optional
        If specified and if the file already exists on destination, then it will be overwritten. By default, the toolbox creates a new file with a new index (eg 'filename_(1).nc').
        Mutually exclusive with ``skip_existing``.
    skip_existing : bool, optional
        If the files already exists where it would be downloaded, then the download is skipped for this file. By default, the toolbox creates a new file with a new index (eg 'filename_(1).nc').
        Mutually exclusive with ``overwrite``.
    service : str, optional
        Force download through one of the available services using the service name among ['arco-geo-series', 'arco-time-series', 'omi-arco', 'static-arco', 'arco-platform-series'] or its short name among ['geoseries', 'timeseries', 'omi-arco', 'static-arco', 'platformseries'].
    request_file : pathlib.Path | str, optional
        Option to pass a file containing the arguments. For more information please refer to the documentation or use option ``--create-template`` from the command line interface for an example template.
    motu_api_request : str, optional
        Option to pass a complete MOTU API request as a string. Caution, user has to replace double quotes " with single quotes ' in the request.
    dry_run : bool, optional
        If True, runs query without downloading data.
    netcdf_compression_level : int, optional
        Specify a compression level to apply on the NetCDF output file. A value of 0 means no compression, and 9 is the highest level of compression available.
    netcdf3_compatible : bool, optional
        Enable downloading the dataset in a netCDF3 compatible format.
    chunk_size_limit : int, default -1
        Limit the size of the chunks in the dask array. Default is set to -1 which behaves similarly to 'chunks=auto' from ``xarray``. Positive integer values and '-1' are accepted. This is an experimental feature.
    raise_if_updating : bool, default False
        If set, raises a :class:`copernicusmarine.DatasetUpdating` error if the dataset is being updated and the subset interval requested overpasses the updating start date of the dataset. Otherwise, a simple warning is displayed.
    platform_ids : list[str], optional
        List of platform IDs to extract. Only available for platform chunked datasets.

    Returns
    -------
    ResponseSubset
        A description of the downloaded data and its destination.
    """  # noqa
    if overwrite:
        if skip_existing:
            raise MutuallyExclusiveArguments("overwrite", "skip_existing")

    if variables is not None:
        _check_type(variables, list, "variables")
    if platform_ids is not None:
        _check_type(platform_ids, list, "platform_ids")

    subset_request = create_subset_request(
        dataset_id=dataset_id,
        dataset_version=dataset_version,
        dataset_part=dataset_part,
        username=username,
        password=password,
        variables=variables,
        minimum_longitude=minimum_longitude,
        maximum_longitude=maximum_longitude,
        minimum_latitude=minimum_latitude,
        maximum_latitude=maximum_latitude,
        minimum_depth=minimum_depth,
        maximum_depth=maximum_depth,
        vertical_axis=vertical_axis,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        minimum_x=minimum_x,
        maximum_x=maximum_x,
        minimum_y=minimum_y,
        maximum_y=maximum_y,
        coordinates_selection_method=coordinates_selection_method,
        output_filename=output_filename,
        file_format=file_format,
        service=service,
        request_file=(pathlib.Path(request_file) if request_file else None),
        output_directory=(
            pathlib.Path(output_directory) if output_directory else None
        ),
        credentials_file=(
            pathlib.Path(credentials_file) if credentials_file else None
        ),
        motu_api_request=motu_api_request,
        overwrite=overwrite,
        skip_existing=skip_existing,
        dry_run=dry_run,
        disable_progress_bar=disable_progress_bar,
        staging=staging,
        netcdf_compression_level=netcdf_compression_level,
        netcdf3_compatible=netcdf3_compatible,
        chunk_size_limit=chunk_size_limit,
        raise_if_updating=raise_if_updating,
        platform_ids=platform_ids,
    )

    return subset_function(
        subset_request=subset_request,
    )


def _check_type(value, expected_type: type, name: str):
    if not isinstance(value, expected_type):
        raise TypeError(f"{name} must be of type {expected_type.__name__}")



import logging
import pathlib

from copernicusmarine.core_functions.credentials_utils import (
    RECOVER_YOUR_CREDENTIALS_MESSAGE,
    copernicusmarine_validate_and_get_user,
    credentials_file_builder,
    try_use_existing_credentials_file,
)

logger = logging.getLogger("copernicusmarine")


def login_function(
    username: str | None,
    password: str | None,
    configuration_file_directory: pathlib.Path,
    force_overwrite: bool,
    check_credentials_valid: bool,
    configuration_file: pathlib.Path | None,
) -> bool:
    if check_credentials_valid:
        logger.info("Checking if credentials are valid.")
        if copernicusmarine_validate_and_get_user(
            configuration_file, username, password
        ):
            return True
        else:
            return False
    if not force_overwrite:
        (
            existing_valid_credentials,
            overwrite,
        ) = try_use_existing_credentials_file(
            configuration_file_directory=configuration_file_directory
        )
    else:
        existing_valid_credentials = False
        overwrite = True
    if existing_valid_credentials and not force_overwrite:
        return True
    if not existing_valid_credentials and not overwrite:
        return False
    if force_overwrite or not existing_valid_credentials:
        credentials_file, has_been_aborted = credentials_file_builder(
            username=username,
            password=password,
            configuration_file_directory=configuration_file_directory,
            force_overwrite=force_overwrite,
        )
        if has_been_aborted:
            logger.info("No configuration file have been modified.")
            return False
        if credentials_file is not None:
            logger.info(f"Credentials file stored in {credentials_file}.")
            return True
        else:
            logger.error("Invalid credentials.")
            logger.info("No configuration file have been modified.")
            logger.info(RECOVER_YOUR_CREDENTIALS_MESSAGE)
    return False



import json
import logging
import pathlib

from copernicusmarine.core_functions.marine_datastore_config import (
    MarineDataStoreConfig,
    get_config_and_check_version_get,
)
from copernicusmarine.core_functions.models import CommandType, ResponseGet
from copernicusmarine.core_functions.request_structure import GetRequest
from copernicusmarine.core_functions.services_utils import (
    RetrievalService,
    get_retrieval_service,
)
from copernicusmarine.core_functions.utils import get_unique_filepath
from copernicusmarine.download_functions.download_original_files import (
    download_original_files,
)

logger = logging.getLogger("copernicusmarine")


def get_function(
    get_request: GetRequest,
    staging: bool,
) -> ResponseGet:
    marine_datastore_config = get_config_and_check_version_get(staging)
    if staging:
        logger.warning(
            "Detecting staging flag for get command. "
            "Data will come from the staging environment."
        )
    return _run_get_request(
        get_request=get_request,
        create_file_list=get_request.create_file_list,
        marine_datastore_config=marine_datastore_config,
        max_concurrent_requests=get_request.max_concurrent_requests,
        disable_progress_bar=get_request.disable_progress_bar,
    )


def _run_get_request(
    get_request: GetRequest,
    create_file_list: str | None,
    marine_datastore_config: MarineDataStoreConfig,
    max_concurrent_requests: int,
    disable_progress_bar: bool,
) -> ResponseGet:
    logger.debug("Checking dataset metadata...")

    retrieval_service: RetrievalService = get_retrieval_service(
        request=get_request,
        command_type=CommandType.GET,
        marine_datastore_config=marine_datastore_config,
    )
    get_request.dataset_url = retrieval_service.uri
    downloaded_files = download_original_files(
        get_request.username,
        get_request,
        max_concurrent_requests,
        disable_progress_bar,
        create_file_list,
    )
    return downloaded_files


def create_get_template() -> None:
    filename = pathlib.Path("get_template.json")
    if filename.exists():
        get_unique_filepath(
            filepath=pathlib.Path("get_template.json"),
        )
    with open(filename, "w") as output_file:
        json.dump(
            {
                "dataset_id": "cmems_mod_ibi_phy-temp_my_0.027deg_P1Y-m",
                "dataset_version": None,
                "dataset_part": None,
                "username": None,
                "password": None,
                "no_directories": False,
                "filter": "*01yav_temp_200[0-2]*",
                "regex": None,
                "output_directory": "copernicusmarine_data",
                "file_list": None,
                "sync": False,
                "sync_delete": False,
                "index_parts": False,
                "disable_progress_bar": False,
                "overwrite": False,
                "log_level": "INFO",
                "dry_run": False,
            },
            output_file,
            indent=4,
        )
    logger.info(f"Template created at: {filename}")



import logging
import pathlib
import shutil
import warnings
from collections import defaultdict
from copy import deepcopy
from datetime import datetime

import pandas as pd
import xarray as xr
from arcosparse import (
    Entity,
    UserConfiguration,
    get_entities,
    subset_and_return_dataframe,
)

from copernicusmarine.catalogue_parser.models import CopernicusMarineService
from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_DISABLE_SSL_CONTEXT,
    COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH,
)
from copernicusmarine.core_functions.exceptions import (
    NotEnoughPlatformMetadata,
    WrongPlatformID,
)
from copernicusmarine.core_functions.models import (  # TimeExtent,
    FileStatus,
    ResponseSubset,
    StatusCode,
    StatusMessage,
    VerticalAxis,
)
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.sessions import TRUST_ENV
from copernicusmarine.core_functions.temporary_path_saver import (
    TemporaryPathSaver,
)
from copernicusmarine.core_functions.utils import (
    construct_query_params_for_marine_data_store_monitoring,
    datetime_parser,
    datetime_to_isoformat,
    datetime_to_timestamp,
    get_unique_directorypath,
    get_unique_filepath,
    timestamp_parser,
)
from copernicusmarine.download_functions.utils import (
    build_filename_from_request,
    get_file_extension,
)
from copernicusmarine.versioner import __version__ as toolbox_version

logger = logging.getLogger("copernicusmarine")

COLUMNS_RENAME = {
    "entity_id": "platform_id",
    "entity_type": "platform_type",
}

COLUMNS_ORDER_DEPTH = [
    "variable",
    "platform_id",
    "platform_type",
    "time",
    "longitude",
    "latitude",
    "depth",
    "pressure",
    "is_depth_from_producer",
    "value",
    "value_qc",
    "institution",
    "doi",
    "product_doi",
]

COLUMNS_ORDER_ELEVATION = deepcopy(COLUMNS_ORDER_DEPTH)
COLUMNS_ORDER_ELEVATION[COLUMNS_ORDER_ELEVATION.index("depth")] = "elevation"

SORTING = {
    "variable": True,
    "platform_id": True,
    "platform_type": True,
    "time": True,
}


def download_sparse(
    username: str,
    subset_request: SubsetRequest,
    metadata_url: str,
    service: CopernicusMarineService,
    axis_coordinate_id_mapping: dict[str, str],
    product_doi: str | None,
    product_id: str | None,
    disable_progress_bar: bool,
) -> ResponseSubset:

    if subset_request.dry_run:
        _, variables, platform_ids = _read_dataframe_sparse(
            username,
            subset_request,
            metadata_url,
            service,
            product_doi,
            disable_progress_bar,
            dry_run=True,
        )
        response = _get_response_subset(
            subset_request,
            variables=variables,
            platform_ids=platform_ids,
            axis_coordinate_id_mapping=axis_coordinate_id_mapping,
        )
        response.status = StatusCode.DRY_RUN
        response.message = StatusMessage.DRY_RUN
        return response

    df, variables, platform_ids = _read_dataframe_sparse(
        username,
        subset_request,
        metadata_url,
        service,
        product_doi,
        disable_progress_bar,
    )
    response = _get_response_subset(
        subset_request,
        variables=variables,
        platform_ids=platform_ids,
        axis_coordinate_id_mapping=axis_coordinate_id_mapping,
    )
    output_path = response.file_path
    if subset_request.skip_existing and output_path.exists():
        response.file_status = FileStatus.IGNORED
        return response
    elif (
        subset_request.overwrite
        and output_path.exists()
        and output_path.is_dir()
    ):
        shutil.rmtree(output_path)

    if subset_request.output_directory:
        subset_request.output_directory.mkdir(parents=True, exist_ok=True)
    if df.empty:
        return response

    if subset_request.file_format == "netcdf":
        user_configuration = _get_user_configuration(username)
        platforms_metadata = {
            entity.entity_id: entity
            for entity in get_entities(metadata_url, user_configuration)
        }
        file_names = _dataframe_to_netcdf_per_platform(
            df=df,
            vertical_axis=subset_request.vertical_axis,
            output_path=output_path,
            platforms_metadata=platforms_metadata,
            subset_request=subset_request,
            product_id=product_id,
            service=service,
            netcdf_compression_level=subset_request.netcdf_compression_level,
            netcdf3_compatible=subset_request.netcdf3_compatible,
        )
        response.file_names = file_names
    else:
        with TemporaryPathSaver(output_path) as tmp_path:
            if subset_request.file_format == "parquet":
                df.to_parquet(tmp_path, index=False)
            else:
                df.to_csv(tmp_path, index=False)

    return response


def read_dataframe_sparse(
    username: str,
    subset_request: SubsetRequest,
    metadata_url: str,
    service: CopernicusMarineService,
    product_doi: str | None,
    disable_progress_bar: bool,
) -> pd.DataFrame:
    df, _, _ = _read_dataframe_sparse(
        username,
        subset_request,
        metadata_url,
        service,
        product_doi,
        disable_progress_bar,
    )
    return df


def _read_dataframe_sparse(
    username: str,
    subset_request: SubsetRequest,
    metadata_url: str,
    service: CopernicusMarineService,
    product_doi: str | None,
    disable_progress_bar: bool,
    dry_run: bool = False,
) -> tuple[pd.DataFrame, list[str], list[str]]:
    """
    Shared function for subset and read dataframe

    Returns also the variables and the platform_ids
    """
    user_configuration = _get_user_configuration(username)
    platforms_metadata = {
        entity.entity_id: entity
        for entity in get_entities(metadata_url, user_configuration)
    }
    if subset_request.platform_ids:
        platform_ids = _get_plaform_ids_to_subset(
            subset_request.platform_ids or [],
            set(platforms_metadata.keys()),
            service,
        )
    else:
        platform_ids = []
    variables = subset_request.variables or [
        variable.short_name for variable in service.variables
    ]
    if dry_run:
        return pd.DataFrame(), variables, platform_ids
    progress_bar_configuration = {
        "disable": disable_progress_bar,
        "bar_format": "{l_bar}{bar}| [{elapsed}<{remaining}]",
    }
    # see https://github.com/pandas-dev/pandas/issues/55928
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=FutureWarning)
        df = subset_and_return_dataframe(
            minimum_latitude=subset_request.minimum_y,
            maximum_latitude=subset_request.maximum_y,
            minimum_longitude=subset_request.minimum_x,
            maximum_longitude=subset_request.maximum_x,
            minimum_elevation=(
                -subset_request.maximum_depth
                if subset_request.maximum_depth is not None
                else None
            ),
            maximum_elevation=(
                -subset_request.minimum_depth
                if subset_request.minimum_depth is not None
                else None
            ),
            minimum_time=(
                subset_request.start_datetime.timestamp()
                if subset_request.start_datetime is not None
                else None
            ),
            maximum_time=(
                subset_request.end_datetime.timestamp()
                if subset_request.end_datetime is not None
                else None
            ),
            variables=variables,
            entities=platform_ids,
            vertical_axis=subset_request.vertical_axis,
            url_metadata=metadata_url,
            user_configuration=user_configuration,
            progress_bar_configuration=progress_bar_configuration,
            columns_rename=COLUMNS_RENAME,
        )

        df = _transform_dataframe(
            df,
            subset_request.vertical_axis,
            platforms_metadata,
            product_doi,
        )
    if df.empty:
        logger.info(
            "No data found for the given parameters. "
            "Please check your request and try again."
        )
    return (
        df,
        variables,
        platform_ids,
    )


def _get_user_configuration(username: str) -> UserConfiguration:
    return UserConfiguration(
        disable_ssl=COPERNICUSMARINE_DISABLE_SSL_CONTEXT == "True",
        trust_env=TRUST_ENV,
        ssl_certificate_path=COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH,
        max_concurrent_requests=20,
        extra_params=construct_query_params_for_marine_data_store_monitoring(
            username
        ),
    )


def _get_plaform_ids_to_subset(
    platform_ids: list[str],
    platforms_metadata_names: set[str],
    retrieval_service: CopernicusMarineService,
) -> list[str]:
    platforms_to_subset = []
    if not platforms_metadata_names:
        raise NotEnoughPlatformMetadata()
    platforms_names_with_types: set[str] = set()
    platforms_without_types_mapping: dict[str, list] = defaultdict(list)
    for platform_name in platforms_metadata_names:
        platform_name_without_type = platform_name.split("___")[0]
        platforms_without_types_mapping[platform_name_without_type].append(
            platform_name
        )
        platforms_names_with_types.add(platform_name)
    for platform_id in platform_ids:
        if platform_id in platforms_names_with_types:
            platforms_to_subset.append(platform_id)
        if platform_id in platforms_without_types_mapping:
            platforms_to_subset.extend(
                platforms_without_types_mapping[platform_id]
            )
    if not platforms_to_subset:
        raise WrongPlatformID(
            platform_ids, retrieval_service.platforms_metadata
        )
    return platforms_to_subset


def _build_filename_and_output_path(
    subset_request: SubsetRequest,
    variables: list[str],
    platform_ids: list[str],
    axis_coordinate_id_mapping: dict[str, str],
) -> tuple[str, pathlib.Path]:
    extension_file = get_file_extension(subset_request.file_format)
    filename = subset_request.output_filename or build_filename_from_request(
        request=subset_request,
        variables=variables,
        platform_ids=platform_ids,
        axis_coordinate_id_mapping=axis_coordinate_id_mapping,
    )
    if subset_request.file_format == "netcdf":
        filename = pathlib.Path(filename).stem
    elif pathlib.Path(filename).suffix != extension_file:
        filename = f"{filename}{extension_file}"
    output_path = pathlib.Path(
        subset_request.output_directory,
        filename,
    )
    if not subset_request.overwrite and not subset_request.skip_existing:
        output_path = (
            get_unique_filepath(output_path)
            if not subset_request.file_format == "netcdf"
            else get_unique_directorypath(output_path)
        )
        filename = output_path.name
    return filename, output_path


def _get_response_subset(
    subset_request: SubsetRequest,
    variables: list[str],
    platform_ids: list[str],
    axis_coordinate_id_mapping: dict[str, str],
) -> ResponseSubset:
    filename, output_path = _build_filename_and_output_path(
        subset_request,
        variables,
        platform_ids,
        axis_coordinate_id_mapping,
    )
    return ResponseSubset(
        file_path=output_path,
        output_directory=subset_request.output_directory,
        filename=filename,
        file_size=None,
        data_transfer_size=None,
        variables=variables,
        # TODO: handle thoses extents maybe opening the dataframe
        coordinates_extent=[],
        status=StatusCode.SUCCESS,
        message=StatusMessage.SUCCESS,
        file_status=FileStatus.DOWNLOADED,
    )


def _transform_dataframe(
    df: pd.DataFrame,
    vertical_axis: VerticalAxis,
    platforms_metadata: dict[str, Entity],
    product_doi: str | None,
) -> pd.DataFrame:
    """
    Transform the dataframe to match the expected format to be consistent with MyOceanPro
    and Copernicus Marine Services.
    """  # noqa
    if df.empty:
        return df
    # Needs to be done before striping the type to the platform_id
    df["institution"] = df["platform_id"].apply(
        lambda x: (
            platforms_metadata[x].institution
            if x in platforms_metadata
            else None
        )
    )
    df["doi"] = df["platform_id"].apply(
        lambda x: (
            platforms_metadata[x].doi if x in platforms_metadata else None
        )
    )

    df["product_doi"] = (
        f"https://doi.org/{product_doi}"
        if product_doi and "https://" not in product_doi
        else product_doi
    )

    # From "platform___type" to "platform" since the type is in
    # column platform_type
    df["platform_id"] = df["platform_id"].str.split("___").str[0]

    df["time"] = df["time"].apply(
        lambda x: datetime_to_isoformat(timestamp_parser(x, unit="s"))
    )

    # Some depth values comes from the arcoification of the data
    # and are calculated from the pressure some others come
    # directly from the producer (ie native/original data)
    df["is_depth_from_producer"] = df["is_approx_elevation"].apply(
        lambda x: 0 if x else 1
    )
    df.drop(columns=["is_approx_elevation"], inplace=True)

    if vertical_axis == "elevation":
        df = df[COLUMNS_ORDER_ELEVATION]
    else:
        df = df[COLUMNS_ORDER_DEPTH]

    df.sort_values(
        by=list(SORTING.keys()),
        ascending=list(SORTING.values()),
        inplace=True,
    )

    df.reset_index(drop=True, inplace=True)
    return df


def _dataframe_to_netcdf_per_platform(
    df: pd.DataFrame,
    vertical_axis: VerticalAxis,
    output_path: pathlib.Path,
    platforms_metadata: dict[str, Entity],
    subset_request: SubsetRequest,
    product_id: str | None,
    service: CopernicusMarineService,
    netcdf_compression_level: int = 0,
    netcdf3_compatible: bool = False,
) -> list[str]:
    vertical_col = "elevation" if vertical_axis == "elevation" else "depth"
    index_cols = ["time", vertical_col]
    aux_cols = [
        "latitude",
        "longitude",
        "pressure",
        "is_depth_from_producer",
    ]
    metadata_cols = [
        "platform_id",
        "platform_type",
        "institution",
        "doi",
        "product_doi",
    ]

    produced_paths: list[str] = []
    with TemporaryPathSaver(output_path, is_directory=True) as tmp_output_path:
        for platform_id, platform_df in df.groupby("platform_id"):
            platform_filename = f"{platform_id}.nc"
            output_path_with_plaform_id = tmp_output_path / platform_filename

            ds = _platform_dataframe_to_dataset(
                platform_df,
                index_cols=index_cols,
                aux_cols=aux_cols,
                metadata_cols=metadata_cols,
            )

            ds = _add_attributes_to_dataset(
                ds=ds,
                platform_df=platform_df,
                vertical_axis=vertical_axis,
                platform_id=str(platform_id),
                metadata_cols=metadata_cols,
                platforms_metadata=platforms_metadata,
                subset_request=subset_request,
                product_id=product_id,
                service=service,
            )

            encoding = _build_netcdf_encoding(ds, netcdf_compression_level)
            netcdf_format = "NETCDF3_CLASSIC" if netcdf3_compatible else None
            engine = "h5netcdf" if not netcdf3_compatible else "netcdf4"

            with TemporaryPathSaver(output_path_with_plaform_id) as tmp_file:
                ds.to_netcdf(
                    tmp_file,
                    encoding=encoding,
                    format=netcdf_format,
                    engine=engine,
                )

            produced_paths.append(platform_filename)
            logger.debug(
                f"Written NetCDF file for platform '{platform_id}': "
                f"{output_path}"
            )

        logger.info(
            f"Produced {len(produced_paths)} NetCDF file(s) "
            f"in {output_path}"
        )
    return produced_paths


def _platform_dataframe_to_dataset(
    platform_df: pd.DataFrame,
    index_cols: list[str],
    aux_cols: list[str],
    metadata_cols: list[str],
) -> xr.Dataset:
    """
    ``depth_level`` is an integer index: for each time step, depth
    observations are ranked by depth. The maximum number of depth
    points across all time steps determines the size of the
    ``depth_level`` dimension.
    """

    obs_df = platform_df.drop(columns=metadata_cols, errors="ignore")

    vertical_col = [c for c in index_cols if c != "time"][0]
    obs_df = obs_df.sort_values(["time", vertical_col])

    pivot = obs_df.pivot_table(
        index=["time", vertical_col],
        columns="variable",
        values=["value", "value_qc"],  # type: ignore[list-item]
        aggfunc="first",
    )
    pivot.columns = [
        f"{var}_qc" if metric == "value_qc" else var
        for metric, var in pivot.columns
    ]

    spatial_and_aux = obs_df.groupby(["time", vertical_col])[aux_cols].first()

    merged = pd.concat([pivot, spatial_and_aux], axis=1)
    merged = merged.reset_index()
    merged["depth_level"] = merged.groupby("time").cumcount()
    merged["time"] = merged["time"].apply(
        lambda x: datetime_to_timestamp(datetime_parser(x), unit="s")
    )
    merged = merged.set_index(["time", "depth_level"])

    ds = merged.to_xarray()
    ds = ds.set_coords("pressure")
    ds = ds.set_coords("latitude")
    ds = ds.set_coords("longitude")
    ds = ds.set_coords(vertical_col)
    ds = ds.set_coords("is_depth_from_producer")
    ds.time.attrs["units"] = "seconds since 1970-01-01T00:00:00Z"
    return ds


def _build_netcdf_encoding(
    ds: xr.Dataset,
    compression_level: int,
) -> dict:
    if compression_level <= 0:
        return {}

    encoding: dict = {}
    for var_name in ds.data_vars:
        encoding[var_name] = {
            "zlib": True,
            "complevel": compression_level,
            "contiguous": False,
            "shuffle": True,
        }
    return encoding


def _add_attributes_to_dataset(
    ds: xr.Dataset,
    platform_df: pd.DataFrame,
    vertical_axis: VerticalAxis,
    platform_id: str,
    metadata_cols: list[str],
    platforms_metadata: dict[str, Entity],
    subset_request: SubsetRequest,
    product_id: str | None,
    service: CopernicusMarineService,
) -> xr.Dataset:
    # global attributes
    first_row = platform_df.iloc[0]
    metadata_info: dict[str, str] = {
        col: str(first_row[col])
        for col in metadata_cols
        if col in first_row.index
        and first_row[col] is not None
        and not pd.isna(first_row[col])  # type: ignore[arg-type]
    }
    if metadata_info.get("platform_id"):
        ds.attrs["platform_code"] = metadata_info["platform_id"]
    if metadata_info.get("institution"):
        ds.attrs["institution"] = metadata_info["institution"]
    if metadata_info.get("doi"):
        ds.attrs["doi"] = metadata_info["doi"]
    if metadata_info.get("product_doi"):
        if "doi" in ds.attrs:
            ds.attrs["doi"] += " " + metadata_info["product_doi"]
        else:
            ds.attrs["doi"] = metadata_info["product_doi"]
    if metadata_info.get("platform_type"):
        platform_id_type = platform_id + "___" + metadata_info["platform_type"]
        platform_metadata = platforms_metadata.get(
            platform_id_type
        ) or platforms_metadata.get(platform_id)
        if platform_metadata and platform_metadata.institution_edmo_code:
            ds.attrs[
                "institution_edmo_code"
            ] = platform_metadata.institution_edmo_code

    ds.attrs["time_coverage_start"] = datetime_to_isoformat(
        datetime_parser(str(platform_df["time"].min()))
    )
    ds.attrs["time_coverage_end"] = datetime_to_isoformat(
        datetime_parser(str(platform_df["time"].max()))
    )
    ds.attrs["last_date_observation"] = ds.attrs["time_coverage_end"]

    if "latitude" in platform_df.columns:
        ds.attrs["geospatial_lat_min"] = float(platform_df["latitude"].min())
        ds.attrs["geospatial_lat_max"] = float(platform_df["latitude"].max())
        ds.attrs["last_latitude_observation"] = ds.sel(
            time=datetime_to_timestamp(
                datetime_parser(ds.attrs["last_date_observation"])
            ),
            method="nearest",
        )["latitude"].values[0]
    if "longitude" in platform_df.columns:
        ds.attrs["geospatial_lon_min"] = float(platform_df["longitude"].min())
        ds.attrs["geospatial_lon_max"] = float(platform_df["longitude"].max())
        ds.attrs["last_longitude_observation"] = ds.sel(
            time=datetime_to_timestamp(
                datetime_parser(ds.attrs["last_date_observation"])
            ),
            method="nearest",
        )["longitude"].values[0]
    if vertical_axis in platform_df.columns:
        ds.attrs["geospatial_vertical_min"] = float(
            platform_df[vertical_axis].min()
        )
        ds.attrs["geospatial_vertical_max"] = float(
            platform_df[vertical_axis].max()
        )
    ds.attrs["references"] = "http://marine.copernicus.eu"
    ds.attrs[
        "license"
    ] = "https://marine.copernicus.eu/user-corner/service-commitments-and-licence"
    ds.attrs["download_date"] = datetime_to_isoformat(datetime.now())
    ds.attrs["copernicusmarine_toolbox_version"] = toolbox_version
    if product_id:
        ds.attrs["product_id"] = product_id
    ds.attrs["dataset_id"] = subset_request.dataset_id
    ds.attrs["Conventions"] = "CF-1.9 ACDD-1.3"
    ds.attrs["title"] = (
        f"Subset of dataset {subset_request.dataset_id} "
        f"with Copernicus Marine Toolbox"
    )
    ds.attrs["contact"] = "servicedesk.cmems@mercator-ocean.eu"

    # time attributes
    ds["time"].attrs["axis"] = "T"
    ds["time"].attrs["standard_name"] = "time"
    # unit is set before

    # depth attributes
    if vertical_axis == "elevation":
        ds["elevation"].attrs["positive"] = "up"
    else:
        ds["depth"].attrs["positive"] = "down"
    ds[vertical_axis].attrs["units"] = "m"
    ds[vertical_axis].attrs["valid_min"] = -12000.0
    ds[vertical_axis].attrs["valid_max"] = 12000.0
    ds[vertical_axis].attrs["standard_name"] = vertical_axis

    # pressure attributes
    ds["pressure"].attrs["units"] = "dbar"
    # ds["pressure"].attrs["axis"] = "Z"
    ds["pressure"].attrs["positive"] = "down"
    ds["pressure"].attrs["standard_name"] = "reference_pressure"

    # latitude and longitude attributes
    ds["latitude"].attrs["units"] = "degrees_north"
    ds["latitude"].attrs["axis"] = "Y"
    ds["latitude"].attrs["valid_min"] = -90.0
    ds["latitude"].attrs["valid_max"] = 90.0
    ds["latitude"].attrs["standard_name"] = "latitude"
    ds["longitude"].attrs["units"] = "degrees_east"
    ds["longitude"].attrs["axis"] = "X"
    ds["longitude"].attrs["valid_min"] = -180.0
    ds["longitude"].attrs["valid_max"] = 180.0
    ds["longitude"].attrs["standard_name"] = "longitude"

    # is depth from producer attributes
    ds["is_depth_from_producer"].attrs["units"] = "1"
    ds["is_depth_from_producer"].attrs["long_name"] = (
        "Flag indicating if the depth value is from the "
        "data producer (1) or calculated from pressure (0)"
    )

    # depth level
    ds["depth_level"].attrs["long_name"] = (
        "Depth level index. "
        "For each time step, depth observations are ranked by depth. "
    )
    ds["depth_level"].attrs["axis"] = "Z"
    ds["depth_level"].attrs["positive"] = "down"
    ds["depth_level"].attrs["units"] = "1"

    # variables
    variables_info = {
        variable.short_name: variable for variable in service.variables
    }
    for variable in ds.data_vars:
        variable = str(variable)
        if "_qc" not in variable:
            variable_info = variables_info[variable]
            ds[variable].attrs["standard_name"] = variable_info.standard_name
            ds[variable].attrs["units"] = variable_info.units
            ds[variable].attrs["ancillary_variable"] = variable + "_qc"
        else:
            variable_without_qc = variable.replace("_qc", "")
            variable_info = variables_info[variable_without_qc]
            ds[variable].attrs[
                "long_name"
            ] = f"Quality control flag for {variable_without_qc}"
            ds[variable].attrs["flags_values"] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            ds[variable].attrs["flags_meanings"] = (
                "no_qc_performed "
                "good_data "
                "probably_good_data "
                "bad_data_that_are_potentially_correctable "
                "bad_data "
                "value_changed "
                "value_below_detection "
                "nominal_value "
                "interpolated_value "
                "missing_value"
            )

    return ds



import logging
import ssl
from typing import Any, Literal

import boto3
import botocore
import botocore.config
import certifi
import requests
import requests.auth
from boto3.s3.transfer import TransferConfig
from requests.adapters import HTTPAdapter, Retry

from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_DISABLE_SSL_CONTEXT,
    COPERNICUSMARINE_HTTPS_RETRIES,
    COPERNICUSMARINE_HTTPS_TIMEOUT,
    COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH,
    COPERNICUSMARINE_TRUST_ENV,
    COPERNICUSMARINE_USE_THREADS,
    PROXY_HTTP,
    PROXY_HTTPS,
)
from copernicusmarine.core_functions.utils import (
    construct_query_params_for_marine_data_store_monitoring,
    create_custom_query_function,
)

logger = logging.getLogger("copernicusmarine")

TRUST_ENV = COPERNICUSMARINE_TRUST_ENV == "True"
PROXIES = {}
if PROXY_HTTP:
    PROXIES["http"] = PROXY_HTTP
if PROXY_HTTPS:
    PROXIES["https"] = PROXY_HTTPS
try:
    HTTPS_TIMEOUT = float(COPERNICUSMARINE_HTTPS_TIMEOUT)
except ValueError:
    HTTPS_TIMEOUT = 60
try:
    HTTPS_RETRIES = int(COPERNICUSMARINE_HTTPS_RETRIES)
except ValueError:
    HTTPS_RETRIES = 5


def get_ssl_context() -> ssl.SSLContext | None:
    if COPERNICUSMARINE_DISABLE_SSL_CONTEXT == "True":
        return None
    if COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH:
        return ssl.create_default_context(
            capath=COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH
        )
    return ssl.create_default_context(cafile=certifi.where())


def get_configured_boto3_session(
    endpoint_url: str,
    operation_type: list[Literal["ListObjectsV2", "HeadObject", "GetObject"]],
    username: str | None = None,
    return_ressources: bool = False,
) -> tuple[Any, Any]:
    config_boto3 = botocore.config.Config(
        signature_version=botocore.UNSIGNED,
        retries={"max_attempts": 10, "mode": "adaptive"},
    )
    s3_session = boto3.Session()
    s3_client = s3_session.client(
        "s3",
        config=config_boto3,
        endpoint_url=endpoint_url,
    )
    for operation in operation_type:
        # Register the botocore event handler for adding custom query params
        # to S3 HEAD and GET requests
        s3_client.meta.events.register(
            f"before-call.s3.{operation}",
            create_custom_query_function(username),
        )
    if not return_ressources:
        return s3_client, None
    s3_resource = boto3.resource(
        "s3",
        config=config_boto3,
        endpoint_url=endpoint_url,
    )
    return s3_client, s3_resource


class ConfiguredBoto3Session:
    def __init__(
        self,
        endpoint_url: str,
        operation_type: list[
            Literal["ListObjectsV2", "HeadObject", "GetObject"]
        ],
        username: str | None = None,
        need_resources: bool = False,
    ):
        self.s3_client, self.s3_resource = get_configured_boto3_session(
            endpoint_url,
            operation_type,
            username,
            return_ressources=need_resources,
        )
        self.use_threads = COPERNICUSMARINE_USE_THREADS

    def close(self):
        self.s3_client.close()
        if self.s3_resource:
            self.s3_resource.meta.client.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def download_file(
        self, bucket_name: str, object_key: str, file_path: str
    ) -> None:
        self.s3_client.download_file(
            bucket_name,
            object_key,
            file_path,
            Config=TransferConfig(use_threads=self.use_threads),
        )

    def get_object(self, bucket_name: str, object_key: str) -> Any:
        response = self.s3_client.get_object(
            Bucket=bucket_name, Key=object_key
        )
        return response


# TODO: add tests
# example: with https://httpbin.org/delay/10 or
# https://medium.com/@mpuig/testing-robust-requests-with-python-a06537d97771
class ConfiguredRequestsSession(requests.Session):
    def __init__(
        self,
        *args,
        timeout: float = HTTPS_TIMEOUT,
        retries: int = HTTPS_RETRIES,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.trust_env = TRUST_ENV
        if COPERNICUSMARINE_DISABLE_SSL_CONTEXT == "True":
            self.verify = False
        else:
            self.verify = (
                COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH or certifi.where()
            )
        self.proxies = PROXIES
        if retries:
            self.mount(
                "https://",
                HTTPAdapter(
                    max_retries=Retry(
                        total=retries,
                        backoff_factor=1,
                        status_forcelist=[
                            408,
                            429,
                            500,
                            502,
                            503,
                            504,
                        ],
                        allowed_methods=None,
                    )
                ),
            )
        self.timeout = timeout

    def request(self, *args, **kwargs):
        kwargs.setdefault("timeout", self.timeout)
        return super().request(*args, **kwargs)


def get_configured_requests_session(
    timeout: float = HTTPS_TIMEOUT, retries: int = HTTPS_RETRIES
) -> requests.Session:
    return ConfiguredRequestsSession(timeout=timeout, retries=retries)


# from https://stackoverflow.com/questions/29931671/making-an-api-call-in-python-with-an-api-that-requires-a-bearer-token # noqa
class BearerAuth(requests.auth.AuthBase):
    """
    Allow to pass the bearer as "auth" argument to the requests.get method and not as a header.

    Hence the headers are not overwritten by the netrc file as stated here:
    https://requests.readthedocs.io/en/latest/user/authentication/#netrc-authentication
    """  # noqa

    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


class JsonParserConnection:
    def __init__(
        self, timeout: float = HTTPS_TIMEOUT, retries: int = HTTPS_RETRIES
    ) -> None:
        self.session = get_configured_requests_session(timeout, retries)

    def get_json_file(self, url: str) -> dict[str, Any]:
        logger.debug(f"Fetching json file at this url: {url}")
        with self.session.get(
            url,
            params=construct_query_params_for_marine_data_store_monitoring(),
            proxies=self.session.proxies,
        ) as response:
            response.raise_for_status()
            return response.json()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()
